# please refer to University of Massachusetts at Amherst
# Computer Science Technical Report TR-06-37
# Bruno Ribeiro, Tao Ye, Don Towsley, Jean Bolot
# Author: Bruno Ribeiro (ribeiro@cs.umass.edu)
#
# WARNING:
# Be careful with the floating point machine precision!
#  The inverse of the Fisher information might strain
#    the floating point precision of your machine ;)
#


library(MASS)

# Sampling probability
p = 0.005
# Maximum flow size
Wmax = 5
# Maximum sample index
samplesIDXMax = Wmax
# Number of original flows (not sampled flows!)
origflows = 1e15

samplingTransform = 1 - (1-p)**(1:Wmax)

#Flow size distribution
origDist = read.table('flowDistNYC21.sorted.plot')
L = pmax(as.matrix(origDist[,3]),1e-250)
normLWmax =  L[1:Wmax]/sum(L[1:Wmax])

# Normalized flow size distribution
theta = (normLWmax*samplingTransform)/sum(normLWmax*samplingTransform)

# Number of sampled flows
samples = origflows * sum(samplingTransform*normLWmax)

print(" Computing B (see IMC 2006 paper)...");
#One monitor sampling matrix without TCP flags
B = matrix(c(1:(samplesIDXMax*Wmax))*0,samplesIDXMax,Wmax)

   for (i in 1:Wmax)
   {
      for (Didx in 1:i)
      {
         B[Didx,i] = dbinom(Didx,i,p)/(1.0 - dbinom(0,i,p))
      }
   }
   
Didx = Wmax

# Sampled flow distribution given theta from trace P[D=d]
d = (B%*%theta)

# Number of parameters of the distribution (unparametric: sDim = Wmax)
sDim = Wmax
interval = 1:(Wmax)

# J is the Fisher information matrix
J = matrix(c(1:(sDim*sDim))*0,sDim,sDim)

print(" Computing Fisher Info Mtx...");
#Fisher information matrix computation
#Refer to the IMC 2006 paper and Cover and Thomas, 1991, pages 326-331

for (Didx in 1:samplesIDXMax)
{
   for (j in 1:Wmax)
   {
       dtheta = B[Didx,interval]  * (1.0/theta[interval]-1) * (theta[interval]**2)
   }
   J = J - dtheta %*% t(dtheta) / d[Didx]
}
 
print(" Starting inverse...");
#The inverse of the Fisher information matrix
# is the Cramer-Rao bound on the covariance matrix of the estimators
#The diagonal of the covariance matrix is the variance of the estimators
# Also, there is a constraint \sum_i \theta_i = 1.
# Addding this constraint with Rb

Jinv = qr.solve(J,tol=1e-20)
F = t((1.0/theta[interval]-1) * (theta[interval]**2))
Rb = Jinv %*% t(F) %*% qr.solve((F%*%Jinv%*%t(F)),tol=1e-20)%*%F%*%Jinv

# Covariance matrix bound on \gamma:
CovMtx = -Jinv + Rb


# Delta method to compute the variance of theta^{\prime}
thTransf = sum(theta / samplingTransform)
gamma = -log(1/theta[interval] - 1)

derivGamma = matrix(1:(length(J))*0,Wmax,Wmax)
E = matrix(1:(length(J))*0,Wmax,Wmax)

for (i in 1:Wmax) {
	aux = (samplingTransform[i]*thTransf)
	derivGamma[,i]  = - exp(-gamma[i]) * theta[interval] * (theta[i]**2) / ( samplingTransform[interval] * aux * thTransf )
	derivGamma[i,i] = (exp(-gamma[i])*theta[i])/aux * ( theta[i] - (theta[i]**2)/aux )
	E[i,i] = (t(derivGamma[,i]) %*% (CovMtx)) %*% derivGamma[,i]  
}

Estd = sqrt(diag(E))
print("With samples:")
print(samples)
print("we have original flow sizes:")

stdDevorigSingle = Estd/sqrt(samples)
print(stdDevorigSingle)

write(stdDevorigSingle, file="stdDevNoTCPFlag.plot",ncolumns=1)


# Delta method to compute the variance of theta
print("With samples:")
print(samples)
print("we have thinned flow sizes:")

gamma = -log(1/theta[interval] - 1)

derivGamma = matrix(1:(length(J))*0,Wmax,Wmax)
E = matrix(1:(length(J))*0,Wmax,Wmax)

for (i in 1:Wmax) {
	derivGamma[i,i]  = exp(-gamma[i]) / ((1 + exp(-gamma[i]))**2)
	E[i,i] = (t(derivGamma[,i]) %*% (CovMtx)) %*% derivGamma[,i]  
}

Estd = sqrt(diag(E))

CRbound = Estd/sqrt(samples)
print(CRbound)

