# please refer to University of Massachusetts at Amherst
# Computer Science Technical Report TR-06-37
# Bruno Ribeiro, Tao Ye, Don Towsley, Jean Bolot
# Author: Bruno Ribeiro (ribeiro@cs.umass.edu)
#
# WARNING:
# Be careful with the floating point machine precision!
#  The inverse of the Fisher information might strain
#    the floating point precision of your machine ;)
#


library(MASS)

# Sampling probability
p = 0.005
# Maximum flow size
Wmax = 50
# Maximum sample index
sampSize = 2*Wmax-1
# Number of original flows (not sampled flows!)
origflows = 1e15

samplingTransform = 1 - (1-p)**(1:Wmax)

# Flow size distribution
origDist = read.table('fakeflowSizeDistribution.plot')
L = pmax(as.matrix(origDist[,1]),1e-250)
normLWmax =  L[1:Wmax]/sum(L[1:Wmax])

# Normalized flow size distribution
theta = (normLWmax*samplingTransform)/sum(normLWmax*samplingTransform)

# Number of sampled flows
samples = origflows * sum(samplingTransform*normLWmax)

print(" Computing B...")
#One monitor sampling matrix
B = matrix(c(1:(sampSize*Wmax))*0,sampSize,Wmax)

# Removed the multiplication by "p" to increase numerical precision 
   for (i in 2:Wmax)
   {
      for (Didx in 1:i)
      {
          B[Didx,i] = p* ((1.0 - p)**(i-Didx))
      }
      # Sample >= Wmax+1 is a sample without SYN
      for (Didx in 1:(i-1))
      {
         B[Didx+Wmax,i] = p*(1.0 - p)**(i-Didx-1) * (1-p) * (i-Didx)
      }
      Didx = 1
      B[Wmax+1,i] = ((1.0 - p)**(i-2))*(i-Didx) * (1-p) 
      B[1,i] = (1.0 - p)**(i-Didx)
   }
   B[1,1] = 1.0
   for (i in 1:Wmax)
   {
      B[,i] = B[,i]/sum(B[,i])
   }

# Sample distribution given theta from trace P[D=d]
d = (B%*%theta)

# Number of parameters of the distribution (unparametric: sDim = Wmax)
sDim = Wmax
interval = 1:(Wmax)

# J is the Fisher information matrix
J = matrix(c(1:(sDim*sDim))*0,sDim,sDim)

print(" Computing Fisher Info Mtx...");
#Fisher information matrix computation
#Refer to the IMC 2006 paper and Cover and Thomas, 1991, pages 326-331

for (Didx in 1:sampSize)
{
   #derivatives that will be used inside the Fisher information matrix computation
   dtheta = B[Didx,interval]  * (1.0/theta[interval]-1) * (theta[interval]**2) 
   J = J - dtheta %*% t(dtheta) / d[Didx]
}
 
print(" Starting inverse...");
#The inverse of the Fisher information matrix
# is the Cramer-Rao bound on the covariance matrix of the estimators
#The diagonal of the covariance matrix is the variance of the estimators
# Also, there is a constraint \sum_i \theta_i = 1.
# Addding this constraint with Rb
Jinv = qr.solve(J,tol=1e-20)
F = t((1.0/theta[interval]-1) * (theta[interval]**2))
Rb = Jinv %*% t(F) %*% qr.solve((F%*%Jinv%*%t(F)),tol=1e-20)%*%F%*%Jinv

# Bound on the gamma covariance matrix:
CovMtx = -Jinv + Rb


samplingTransform = 1 - (1-p)**(1:Wmax)
thTransf = sum(theta / samplingTransform)
gamma = -log(1/theta[interval] - 1)

# Delta method to compute the variance of theta^{\prime}
derivGamma = matrix(1:(length(J))*0,Wmax,Wmax)
E = matrix(1:(length(J))*0,Wmax,Wmax)

for (i in 1:Wmax) {
	aux = (samplingTransform[i]*thTransf)
	derivGamma[,i]  = - exp(-gamma[i]) * theta[interval] * (theta[i]**2) / ( samplingTransform[interval] * aux * thTransf )
	derivGamma[i,i] = (exp(-gamma[i])*theta[i])/aux * ( theta[i] - (theta[i]**2)/aux )
	E[i,i] = (t(derivGamma[,i]) %*% (CovMtx)) %*% derivGamma[,i]  
}



Estd = sqrt(diag(E))
print("With samples:")
print(samples)
print("we have original flow sizes:")

stdDevorigSingle = Estd/sqrt(samples)
print(stdDevorigSingle)

write(stdDevorigSingle, file="stdDevTCPSeqMixsingle.plot",ncolumns=1)

print("With samples:")
print(samples)
print("we have thinned flow sizes:")

# Delta method to compute the variance of theta
gamma = -log(1/theta[interval] - 1)

derivGamma = matrix(1:(length(J))*0,Wmax,Wmax)
E = matrix(1:(length(J))*0,Wmax,Wmax)

for (i in 1:Wmax) {
	derivGamma[i,i]  = exp(-gamma[i]) / ((1 + exp(-gamma[i]))**2)
	E[i,i] = (t(derivGamma[,i]) %*% (CovMtx)) %*% derivGamma[,i]  
}

Estd = sqrt(diag(E))

CRbound = Estd/sqrt(samples)
print(CRbound)
