# please refer to University of Massachusetts at Amherst
# Computer Science Technical Report TR-06-37
# Bruno Ribeiro, Tao Ye, Don Towsley, Jean Bolot
# Author: Bruno Ribeiro (ribeiro@cs.umass.edu)
#
# WARNING:
# Be careful with the floating point machine precision!
#  The inverse of the Fisher information might strain
#    the floating point precision of your machine ;)
#


library(MASS)

# Sampling probability
p = 1/(64*2)
# Maximum flow size
Wmax = 50
# Maximum sample index
sampSize = 2*Wmax-1
# Number of original flows (not sampled flows!)
origflows = 1e15
# Number of monitors
monitors = 2

samplingTransform = 1 - ((1-p)**(1:Wmax))**monitors

#Flow size distribution
origDist = read.table('fakeflowSizeDistribution.plot')
L = pmax(as.matrix(origDist[,1]),1e-250)
normLWmax =  L[1:Wmax]/sum(L[1:Wmax])

# Normalized flow size distribution
theta = (normLWmax*samplingTransform)/sum(normLWmax*samplingTransform)

# Number of sampled flows
samples = origflows * sum(samplingTransform*normLWmax)

print(" Computing B...")
#K monitor sampling matrix
B = matrix(c(1:(sampSize*Wmax))*0,sampSize,Wmax)

# Removed the multiplication by "p" to increase numerical precision 
   for (i in 2:Wmax)
   {
      for (Didx in 1:i)
      {
          B[Didx,i] = (1-(1-p)**monitors) * (((1-p)**(i-Didx))**monitors)
      }
      # Sample >= Wmax+1 is a sample without SYN
      for (Didx in 1:(i-1))
      {
         B[Didx+Wmax,i] = (1-(1-p)**monitors) * (((1-p)**(i-Didx))**monitors) * (i-Didx)
      }
      Didx = 1
      B[Wmax+1,i] = (((1-p)**(i-1))**monitors)*(i-Didx) 
      B[1,i] = ((1-p)**(i-Didx))**monitors
   }
   B[1,1] = 1.0
   for (i in 1:Wmax)
   {
      B[,i] = B[,i]/sum(B[,i])
  }

# Sample distribution given theta from trace P[D=d]
d = (B%*%theta)

# Number of parameters of the distribution (unparametric: sDim = Wmax)
sDim = Wmax
interval = 1:(Wmax)

# J is the Fisher information matrix
J = matrix(c(1:(sDim*sDim))*0,sDim,sDim)

print(" Computing Fisher Info Mtx...");
#Fisher information matrix computation
#Refer to the IMC 2006 paper and Cover and Thomas, 1991, pages 326-331

for (Didx in 1:sampSize)
{
   #derivatives that will be used inside the Fisher information matrix computation
   dtheta = B[Didx,interval]  * (1.0/theta[interval]-1) * (theta[interval]**2) 
   J = J - dtheta %*% t(dtheta) / d[Didx]
}
 
print(" Starting inverse...");
#The inverse of the Fisher information matrix
# is the Cramer-Rao bound on the covariance matrix of the estimators
#The diagonal of the covariance matrix is the variance of the estimators
# Also, there is a constraint \sum_i \theta_i = 1.
# Addding this constraint with Rb
Jinv = qr.solve(J,tol=1e-20)
F = t((1.0/theta[interval]-1) * (theta[interval]**2))
Rb = Jinv %*% t(F) %*% qr.solve((F%*%Jinv%*%t(F)),tol=1e-20)%*%F%*%Jinv

# Covariance matrix bound on \gamma:
CovMtx = -Jinv + Rb


# Delta method to compute the variance of theta^{\prime}
thTransf = sum(theta / samplingTransform)
gamma = -log(1/theta[interval] - 1)

derivGamma = matrix(1:(length(J))*0,Wmax,Wmax)
E = matrix(1:(length(J))*0,Wmax,Wmax)

for (i in 1:Wmax) {
	aux = (samplingTransform[i]*thTransf)
	derivGamma[,i]  = - exp(-gamma[i]) * theta[interval] * (theta[i]**2) / ( samplingTransform[interval] * aux * thTransf )
	derivGamma[i,i] = (exp(-gamma[i])*theta[i])/aux * ( theta[i] - (theta[i]**2)/aux )
	E[i,i] = (t(derivGamma[,i]) %*% (CovMtx)) %*% derivGamma[,i]
}


Estd = sqrt(diag(E))

print("With samples:")
print(samples)
print("we have original flow sizes:")
origMult = Estd/sqrt(samples)
print(origMult)

write(origMult, file="origMult_TCPAllflags.plot",ncolumns=1)
